@@ -24,6 +24,7 @@ import platform
 from argparse import ArgumentParser, RawDescriptionHelpFormatter, REMAINDER
 from base64 import b64encode
 from hashlib import sha256
+from itertools import count
 from os import getenv, makedirs, listdir
 from os.path import join as path_join, normpath, realpath, isdir, isfile, getsize
 from random import randint
@@ -49,12 +50,7 @@ CORE_DIR_FORMAT = "core-%d"
 READ_REPLICAS_DIR = "read-replicas"
 READ_REPLICA_DIR_FORMAT = "read-replica-%d"
 
-INITIAL_DISCOVERY_PORT = 5000
-INITIAL_TRANSACTION_PORT = 6000
-INITIAL_RAFT_PORT = 7000
-INITIAL_BOLT_PORT = 7687
-INITIAL_HTTP_PORT = 7474
-INITIAL_HTTPS_PORT = 6474
+DEFAULT_INITIAL_PORT = 20000
 
 
 def bstr(s):
@@ -168,7 +164,7 @@ class Downloader(object):
         return package
 
 
-def wait_for_server(host, port, timeout=30):
+def wait_for_server(host, port, timeout=90):
     port = port or 7474
     running = False
     t = 0
@@ -187,7 +183,7 @@ def wait_for_server(host, port, timeout=30):
 
 
 def hex_bytes(data):
-    return b"".join(b"%02X" % b for b in bytearray(data))
+    return "".join("%02X" % b for b in bytearray(data))
 
 
 def user_record(user, password):
@@ -195,7 +191,7 @@ def user_record(user, password):
     m = sha256()
     m.update(salt)
     m.update(bstr(password))
-    return b"%s:SHA-256,%s,%s:" % (bstr(user), hex_bytes(m.digest()), hex_bytes(salt))
+    return bstr("%s:SHA-256,%s,%s:" % (user, hex_bytes(m.digest()), hex_bytes(salt)))
 
 
 class Controller(object):
@@ -226,10 +222,10 @@ class Controller(object):
         self.home = home
         self.verbosity = verbosity
 
-    def start(self, wait):
+    def start(self, wait=True):
         raise NotImplementedError("Not yet supported for this platform")
 
-    def stop(self, kill):
+    def stop(self, kill=False):
         raise NotImplementedError("Not yet supported for this platform")
 
     def create_user(self, user, password):
@@ -281,7 +277,7 @@ class Controller(object):
 
         if self._neo4j_admin_available():
             neo4j_admin_path = path_join(self.home, "bin", self._neo4j_admin_script_name())
-            output = check_output([neo4j_admin_path, "set-initial-password", password])
+            output = _invoke([neo4j_admin_path, "set-initial-password", password])
             live_users_detected = "password was not set" in output.lower()
         else:
             if self._auth_file_exists():
@@ -295,7 +291,7 @@ class Controller(object):
     def _neo4j_admin_available(self):
         try:
             neo4j_admin_path = path_join(self.home, "bin", self._neo4j_admin_script_name())
-            output = check_output([neo4j_admin_path, "help"])
+            output = _invoke([neo4j_admin_path, "help"])
             return "set-initial-password" in output
         except CalledProcessError:
             return False
@@ -326,21 +322,21 @@ class UnixController(Controller):
 
     def start(self, wait=True):
         http_uri, bolt_uri = config.extract_http_and_bolt_uris(self.home)
-        check_output([path_join(self.home, "bin", "neo4j"), "start"])
+        _invoke([path_join(self.home, "bin", "neo4j"), "start"])
         if wait:
             wait_for_server(http_uri.hostname, http_uri.port)
         return InstanceInfo(http_uri, bolt_uri, self.home)
 
-    def stop(self, kill):
+    def stop(self, kill=False):
         if kill:
-            output = check_output([path_join(self.home, "bin", "neo4j"), "status"])
+            output = _invoke([path_join(self.home, "bin", "neo4j"), "status"])
             if output.startswith("Neo4j is running"):
                 pid = output.split(" ")[-1].strip()
-                check_output(["kill", "-9", pid])
+                _invoke(["kill", "-9", pid])
             else:
                 raise RuntimeError("Neo4j is not running")
         else:
-            check_output([path_join(self.home, "bin", "neo4j"), "stop"])
+            _invoke([path_join(self.home, "bin", "neo4j"), "stop"])
 
     @classmethod
     def _neo4j_admin_script_name(cls):
@@ -364,16 +360,16 @@ class WindowsController(Controller):
 
     def start(self, wait=True):
         http_uri, bolt_uri = config.extract_http_and_bolt_uris(self.home)
-        check_output([path_join(self.home, "bin", "neo4j.bat"), "install-service"])
-        check_output([path_join(self.home, "bin", "neo4j.bat"), "start"])
+        _invoke([path_join(self.home, "bin", "neo4j.bat"), "install-service"])
+        _invoke([path_join(self.home, "bin", "neo4j.bat"), "start"])
         if wait:
             wait_for_server(http_uri.hostname, http_uri.port)
         return InstanceInfo(http_uri, bolt_uri, self.home)
 
-    def stop(self, kill):
+    def stop(self, kill=False):
         if kill:
             windows_service_name = config.extract_windows_service_name(self.home)
-            sc_output = check_output(["sc", "queryex", windows_service_name])
+            sc_output = _invoke(["sc", "queryex", windows_service_name])
             pid = None
             for line in sc_output.splitlines():
                 line = line.strip()
@@ -383,11 +379,11 @@ class WindowsController(Controller):
             if pid is None:
                 raise RuntimeError("Unable to get PID")
 
-            check_output(["taskkill", "/f", "/pid", pid])
+            _invoke(["taskkill", "/f", "/pid", pid])
         else:
-            check_output([path_join(self.home, "bin", "neo4j.bat"), "stop"])
+            _invoke([path_join(self.home, "bin", "neo4j.bat"), "stop"])
 
-        check_output([path_join(self.home, "bin", "neo4j.bat"), "uninstall-service"])
+        _invoke([path_join(self.home, "bin", "neo4j.bat"), "uninstall-service"])
 
     @classmethod
     def _neo4j_admin_script_name(cls):
@@ -398,12 +394,14 @@ class Cluster:
     def __init__(self, path):
         self.path = path
 
-    def install(self, version, core_count, read_replica_count, verbose=False):
+    def install(self, version, core_count, read_replica_count, initial_port, password, verbose=False):
         try:
             package = _create_controller().download("enterprise", version, self.path, verbose=verbose)
+            port_gen = count(initial_port)
 
-            initial_discovery_members = self._install_cores(self.path, package, core_count)
-            self._install_read_replicas(self.path, package, initial_discovery_members, core_count, read_replica_count)
+            initial_discovery_members = self._install_cores(self.path, package, core_count, port_gen)
+            self._install_read_replicas(self.path, package, initial_discovery_members, read_replica_count, port_gen)
+            self._set_initial_password(password)
 
             return realpath(self.path)
 
@@ -426,11 +424,11 @@ class Cluster:
     def stop(self, kill):
         self._foreach_cluster_member(self._cluster_member_kill if kill else self._cluster_member_stop)
 
-    def set_initial_password(self, password):
+    def _set_initial_password(self, password):
         self._foreach_cluster_member(lambda path: self._cluster_member_set_initial_password(path, password))
 
     @classmethod
-    def _install_cores(cls, path, package, core_count):
+    def _install_cores(cls, path, package, core_count, port_generator):
         discovery_listen_addresses = []
         transaction_listen_addresses = []
         raft_listen_addresses = []
@@ -439,12 +437,12 @@ class Cluster:
         https_listen_addresses = []
 
         for core_idx in range(0, core_count):
-            discovery_listen_addresses.append(_localhost(INITIAL_DISCOVERY_PORT + core_idx))
-            transaction_listen_addresses.append(_localhost(INITIAL_TRANSACTION_PORT + core_idx))
-            raft_listen_addresses.append(_localhost(INITIAL_RAFT_PORT + core_idx))
-            bolt_listen_addresses.append(_localhost(INITIAL_BOLT_PORT + core_idx))
-            http_listen_addresses.append(_localhost(INITIAL_HTTP_PORT + core_idx))
-            https_listen_addresses.append(_localhost(INITIAL_HTTPS_PORT + core_idx))
+            discovery_listen_addresses.append(_localhost(next(port_generator)))
+            transaction_listen_addresses.append(_localhost(next(port_generator)))
+            raft_listen_addresses.append(_localhost(next(port_generator)))
+            bolt_listen_addresses.append(_localhost(next(port_generator)))
+            http_listen_addresses.append(_localhost(next(port_generator)))
+            https_listen_addresses.append(_localhost(next(port_generator)))
 
         initial_discovery_members = ",".join(discovery_listen_addresses)
 
@@ -470,20 +468,16 @@ class Cluster:
         return initial_discovery_members
 
     @classmethod
-    def _install_read_replicas(cls, path, package, initial_discovery_members, core_count, read_replica_count):
-        first_bolt_port = INITIAL_BOLT_PORT + core_count
-        first_http_port = INITIAL_HTTP_PORT + core_count
-        first_https_port = INITIAL_HTTPS_PORT + core_count
-
+    def _install_read_replicas(cls, path, package, initial_discovery_members, read_replica_count, port_generator):
         controller = _create_controller()
         for read_replica_idx in range(0, read_replica_count):
             read_replica_dir = READ_REPLICA_DIR_FORMAT % read_replica_idx
             read_replica_path = path_join(path, READ_REPLICAS_DIR, read_replica_dir)
             read_replica_home = controller.extract(package, read_replica_path)
 
-            bolt_listen_address = _localhost(first_bolt_port + read_replica_idx)
-            http_listen_address = _localhost(first_http_port + read_replica_idx)
-            https_listen_address = _localhost(first_https_port + read_replica_idx)
+            bolt_listen_address = _localhost(next(port_generator))
+            http_listen_address = _localhost(next(port_generator))
+            https_listen_address = _localhost(next(port_generator))
 
             read_replica_config = config.for_read_replica(initial_discovery_members, bolt_listen_address,
                                                           http_listen_address, https_listen_address)
@@ -646,8 +640,7 @@ def cluster():
     sub_commands_with_description = {
         "install": "Download, extract and configure causal cluster",
         "start": "Start the causal cluster located at the given path",
-        "stop": "Stop the causal cluster located at the given path",
-        "set-initial-password": "Sets the initial password of the initial admin user ('neo4j') for all cluster members",
+        "stop": "Stop the causal cluster located at the given path"
     }
 
     subparsers = parser.add_subparsers(title="available sub-commands", dest="command",
@@ -667,6 +660,12 @@ def cluster():
                                 help="Number of core members in the cluster (default 3)")
     parser_install.add_argument("-r", "--read-replicas", default="0", dest="read_replica_count",
                                 help="Number of read replicas in the cluster (default 0)")
+    parser_install.add_argument("-i", "--initial-port", default=str(DEFAULT_INITIAL_PORT),
+                                dest="initial_port", help="Initial port number for all used ports on all cluster "
+                                                          "members. Each next port will simply be an increment of "
+                                                          "the previous one (default %d)" % DEFAULT_INITIAL_PORT)
+    parser_install.add_argument("-p", "--password", required=True,
+                                help="initial password of the initial admin user ('neo4j') for all cluster members")
     parser_install.add_argument("path", nargs="?", default=".", help="download destination path (default: .)")
 
     parser_start = subparsers.add_parser("start", epilog=see_download_command,
@@ -676,8 +675,8 @@ def cluster():
                                          "  neoctrl-cluster start $HOME/cluster/",
                                          formatter_class=RawDescriptionHelpFormatter)
 
-    parser_start.add_argument("-t", "--timeout", default="120", dest="timeout",
-                              help="startup timeout in seconds (default: 120)")
+    parser_start.add_argument("-t", "--timeout", default="180", dest="timeout",
+                              help="startup timeout in seconds (default: 180)")
 
     parser_start.add_argument("path", nargs="?", default=".", help="causal cluster location path (default: .)")
 
@@ -692,20 +691,7 @@ def cluster():
                              help="forcefully kill all instances in the cluster")
     parser_stop.add_argument("path", nargs="?", default=".", help="causal cluster location path (default: .)")
 
-    parser_set_initial_pwd = subparsers.add_parser("set-initial-password", epilog=see_download_command,
-                                                   description=
-                                                   sub_commands_with_description["set-initial-password"] +
-                                                   "\r\n\r\nexample:\r\n"
-                                                   "  neoctrl-set-initial-password newPassword $HOME/cluster/",
-                                                   formatter_class=RawDescriptionHelpFormatter)
-
-    parser_set_initial_pwd.add_argument("password", help="password for the admin user")
-
-    parser_set_initial_pwd.add_argument("path", nargs="?", default=".",
-                                        help="causal cluster location path (default: .)")
-
     parsed = parser.parse_args()
-
     _execute_cluster_command(parsed)
 
 
@@ -722,18 +708,18 @@ def _execute_cluster_command(parsed):
     command = parsed.command
     cluster_ctrl = Cluster(parsed.path)
     if command == "install":
-        core_count = int(parsed.core_count)
-        read_replica_count = int(parsed.read_replica_count)
-        path = cluster_ctrl.install(parsed.version, core_count, read_replica_count, parsed.verbose)
+        core_count = _parse_int(parsed.core_count, "core count")
+        read_replica_count = _parse_int(parsed.read_replica_count, "read replica count")
+        initial_port = _parse_int(parsed.initial_port, "initial port")
+        path = cluster_ctrl.install(parsed.version, core_count, read_replica_count, initial_port, parsed.password,
+                                    parsed.verbose)
         print(path)
     elif command == "start":
-        startup_timeout = int(parsed.timeout)
+        startup_timeout = _parse_int(parsed.timeout, "timeout")
         cluster_info = cluster_ctrl.start(startup_timeout)
         print(cluster_info)
     elif command == "stop":
         cluster_ctrl.stop(parsed.kill)
-    elif command == "set-initial-password":
-        cluster_ctrl.set_initial_password(parsed.password)
     else:
         raise RuntimeError("Unknown command %s" % command)
 
@@ -871,19 +857,37 @@ def test():
         controller.set_user_role("neotest", "admin")
         try:
             controller.start()
-            exit_status = call(parsed.command, *parsed.args)
+            exit_status = call([parsed.command] + parsed.args)
+        except OSError:
+            raise RuntimeError("Unable to run command %r with "
+                               "arguments %r" % (parsed.command, parsed.args))
         finally:
             controller.stop()
-        print()
+        print("")
         if exit_status != 0:
             break
     exit(exit_status)
 
 
+def _invoke(command):
+    try:
+        return check_output(command)
+    except CalledProcessError as error:
+        print("Command failed.\r\nError code: %s\r\nOutput:\r\n%s\n\r" % (str(error.returncode), error.output))
+        raise
+
+
 def _create_controller(path=None):
     return WindowsController(path) if platform.system() == "Windows" else UnixController(path)
 
 
+def _parse_int(string_value, message):
+    try:
+        return int(string_value)
+    except ValueError:
+        raise RuntimeError("Int value expected for %s but received '%s'" % (message, string_value))
+
+
 def _localhost(port):
     return "localhost:%d" % port
 
